# pylint: disable=F,W
from .attributes import Attribute
from .attributes import set_read_memoization
from .attributes.blobs import Blob
from .attributes.blobs import BlobAttribute
from .attributes.collections import DictAttribute
from .attributes.collections import LargeDictAttribute
from .attributes.collections import LargeListAttribute
from .attributes.collections import ListAttribute
from .attributes.collections import SetAttribute
from .attributes.collections import SmallDictAttribute
from .attributes.collections import SmallListAttribute
from .attributes.counter import CounterAttribute
from .attributes.enum import EnumAttribute
from .attributes.files import FileAttribute
from .attributes.files import FileObject
from .attributes.files import FilesGarbageCollector
from .attributes.files import set_files_root_dir
from .attributes.index import Index
from .attributes.models import IndexedModelAttribute
from .attributes.models import InlineModelAttribute
from .attributes.models import ModelAttribute
from .attributes.models import ReverseModelAttribute
from .attributes.password import PasswordAttribute
from .attributes.simples import BooleanAttribute
from .attributes.simples import DateAttribute
from .attributes.simples import DateTimeAttribute
from .attributes.simples import FloatAttribute
from .attributes.simples import IntegerAttribute
from .attributes.simples import SimpleAttribute
from .attributes.simples import StringAttribute
from .attributes.simples import StringUUIDAttribute
from .attributes.simples import TimeAttribute
from .attributes.simples import UUIDAttribute
from .constants import ASC
from .constants import DESC
from .databases import connection
from .databases import Database
from .exceptions import EmptyQuerySetUnpackException
from .exceptions import InvalidFilterException
from .exceptions import InvalidIndexException
from .exceptions import InvalidOrderException
from .exceptions import ModelObjectNotFoundException
from .exceptions import NotConnectedException
from .exceptions import ObjectNotFoundException
from .exceptions import PrimaryKeyException
from .exceptions import QuerySetUnpackException
from .exceptions import SameNameForTableException
from .exceptions import SherafException
from .exceptions import UniqueIndexException
from .health import check_health
from .health import print_health
from .models import AttributeModel
from .models import IntIndexedIntAttributesModel
from .models import IntIndexedModel
from .models import IntIndexedNamedAttributesModel
from .models import IntOrderedIndexedModel
from .models import IntOrderedNamedAttributesModel
from .models import Model
from .models import UUIDIndexedDatedNamedAttributesModel
from .models import UUIDIndexedModel
from .models import UUIDIndexedNamedAttributesModel
from .models.attributes import DatedNamedAttributesModel
from .models.attributes import IntAttributesModel
from .models.attributes import NamedAttributesModel
from .models.base import BaseModel
from .models.indexation import BaseIndexedModel
from .models.indexation import IndexedModel
from .models.inline import InlineModel
from .queryset import QuerySet
from .transactions import attempt
from .transactions import commit
from .version import __version__
from .version import __version_info__



import math
import multiprocessing
import sys

import click
import sheraf
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import TimeRemainingColumn
from rich.traceback import install
from sheraf.health.utils import discover_models


install(show_locals=True)


@click.group()
@click.argument("uri")
@click.pass_context
def cli(ctx, uri):
    sys.path.append(".")
    ctx.db = sheraf.Database(uri)


@cli.result_callback()
@click.pass_context
def after_cli(ctx, result, **kwargs):
    ctx.db.close()


@cli.command()
@click.argument("models", nargs=-1)
def check(models):
    with sheraf.connection():
        sheraf.print_health(*models)


@cli.command()
@click.argument("models")
@click.option(
    "--index",
    help="The name of the indexes to rebuild. If not provided all the indexes will be rebuilt.",
    multiple=True,
)
@click.option(
    "--batch-size",
    help="The number of elements to iterate between two transactions savepoints or commits.",
    default=1000,
    type=int,
)
@click.option(
    "--commit/--no-commit",
    help="Make a real commit for each batch. Defaults to False.",
    default=False,
    is_flag=True,
)
@click.option(
    "--reset/--no-reset",
    help="Delete the whole index before rebuilding it. Defaults to True.",
    default=True,
    is_flag=True,
)
@click.option(
    "--fork/--no-fork",
    help="Computes each batch in a different process. Implies --commit. Defaults to False.",
    default=False,
    is_flag=True,
)
@click.option(
    "--start",
    help="The indice of the first element to reset.",
    default=None,
    type=int,
)
@click.option(
    "--end",
    help="The indice of the last element to reset.",
    default=None,
    type=int,
)
def rebuild(models, index, batch_size, commit, reset, fork, start, end):
    with Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        TextColumn("{task.completed}", justify="right"),
        TextColumn("/"),
        TextColumn("{task.total}"),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeRemainingColumn(),
        transient=True,
    ) as progress:
        models = discover_models(models)

        if not fork:
            with sheraf.connection(commit=True) as conn:
                for _, model in models:
                    total = min(end or math.inf, model.count()) - (start or 0)
                    task = progress.add_task(model.table, total=total)

                    def callback(i, m):
                        if i and i % batch_size == 0:
                            if commit:
                                conn.transaction_manager.commit()
                                conn.cacheGC()
                            else:
                                conn.transaction_manager.savepoint(True)
                        progress.update(task, advance=1)

                    model.index_table_rebuild(
                        *index, callback=callback, reset=reset, start=start, end=end
                    )

        else:

            def process(uri, model, index, start, end):
                if uri:
                    sheraf.Database(uri)

                with sheraf.connection(commit=True):
                    model.index_table_rebuild(*index, reset=False, start=start, end=end)

            for _, model in models:
                with sheraf.connection(commit=True):
                    total = min(end or math.inf, model.count()) - (start or 0)
                    task = progress.add_task(model.table, total=total)
                    uri = sheraf.Database.get().uri
                    nb_batches = math.ceil(total / batch_size)

                    if reset:
                        model.index_table_reset(index)

                for i in range(nb_batches):
                    start = batch_size * i
                    end = min(batch_size * (i + 1), total)
                    p = multiprocessing.Process(
                        target=process,
                        args=(
                            uri,
                            model,
                            index,
                            start,
                            end,
                        ),
                    )
                    p.start()
                    p.join()
                    progress.update(task, advance=end - start)



import types

import sheraf.attributes

from ..types import SmallDict


class BaseModelMetaclass(type):
    """
    Internal metaclass.
    Contains the mapping of attribute names with their corresponding data (of type :class:`~sheraf.attributes.Attribute`)
    """

    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)
        klass.attributes = {}
        klass.cb_creation = []
        klass.cb_deletion = []

        for name, attr in attrs.items():
            if not isinstance(attr, sheraf.attributes.Attribute):
                continue

            try:
                attr.attribute_name = klass.attribute_id(name, attr)
            except NotImplementedError:
                continue

            klass.attributes[name] = attr

        for _base in bases:
            base_attributes = {}
            base_attributes.update(_base.__dict__.get("attributes", {}))
            base_attributes.update(_base.__dict__)
            for name, attr in base_attributes.items():
                if not isinstance(attr, sheraf.attributes.Attribute):
                    continue

                if name in klass.attributes:
                    continue

                try:
                    attr.attribute_name = klass.attribute_id(name, attr)
                except NotImplementedError:
                    continue

                klass.attributes[name] = attr

        return klass


class BaseModel(metaclass=BaseModelMetaclass):
    """
    :class:`~sheraf.models.base.BaseModel` is the base class for every other model classes.
    This is where the attribute reading and writing are handled.

    Models can be used as dictionaries:

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection(): # doctest: +SKIP
    ...     dict(Cowboy.create(name="George Abitbol"))
    {'name': 'George Abitbol', '_creation': ...}
    """

    attributes = {}
    mapping = None
    default_mapping = SmallDict

    @classmethod
    def create(cls, default=None, *args, **kwargs):
        """Create a model instance.

        :param default: The data structure that will be used to store the instance state.
        :param \\*\\*kwargs: Any model attribute can be initialized with the matching keyword.

        :return: The newly created instance.

        >>> class Cowboy(sheraf.Model):
        ...    table = "cowboy"
        ...    name = sheraf.SimpleAttribute(default="John Doe")
        ...
        >>> with sheraf.connection(commit=True):
        ...    cowboy = Cowboy.create()
        ...    assert "John Doe" == cowboy.name
        ...
        ...    cowboy = Cowboy.create(name="George Abitbol")
        ...    assert "George Abitbol" == cowboy.name
        ...
        ...    Cowboy.create(this_attribute_does_not_exist="something") #  raises a TypeError
        Traceback (most recent call last):
            ...
        TypeError: TypeError: create() got an unexpected keyword argument 'this_attribute_does_not_exist'

        The function can also create sub-instances recursively:

        >>> class Horse(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "cowboy"
        ...     name = sheraf.SimpleAttribute(default="John Doe")
        ...     horse = sheraf.InlineModelAttribute(Horse)
        ...
        >>> with sheraf.connection():
        ...     cowboy = Cowboy.create(name="George Abitbol", horse={"name": "Jolly Jumper"})
        ...     cowboy.horse.name
        'Jolly Jumper'
        """

        instance = None
        try:
            mapping = (default or cls.default_mapping)()
            instance = cls._decorate(mapping)
            yield_callbacks = cls.call_callbacks(cls.cb_creation, instance)
            instance.initialize(**kwargs)
            cls.call_callbacks_again(yield_callbacks)
            return instance
        except Exception:
            try:
                if instance:
                    instance.delete()
            except:
                pass
            raise

    def delete(self):
        """Delete the current model instance.

        >>> class MyModel(sheraf.Model):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...    m = MyModel.create()
        ...    assert m == MyModel.read(m.id)
        ...    m.delete()
        ...    m.read(m.id)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id '...' not found in MyModel
        """

        cls = self.__class__
        yield_callbacks = cls.call_callbacks(cls.cb_deletion, self)
        for attr_name in self.attributes.keys():
            delattr(self, attr_name)
        cls.call_callbacks_again(yield_callbacks)

    def initialize(self, **kwargs):
        for attribute, value in kwargs.items():
            if attribute not in self.attributes:
                raise TypeError(
                    "TypeError: create() got an unexpected keyword argument '{}'".format(
                        attribute
                    )
                )
            self.__setattr__(attribute, value)

        for name, attribute in self.attributes.items():
            if (
                not attribute.lazy
                and name not in kwargs
                and not attribute.is_created(self)
            ):
                self.__setattr__(name, attribute.create(self))

    @staticmethod
    def call_callbacks(callbacks, *args, **kwargs):
        yield_callbacks = []
        for callback in callbacks:
            res = callback(*args, **kwargs)
            if isinstance(res, types.GeneratorType):
                try:
                    next(res)
                except StopIteration:
                    pass
                yield_callbacks.append(res)
        return yield_callbacks

    @staticmethod
    def call_callbacks_again(callbacks):
        for callback in callbacks:
            try:
                next(callback)
            except StopIteration:
                pass

    @classmethod
    def on_creation(cls, *args, **kwargs):
        """
        Decorator for callbacks to call on an instance creation. The callback
        will be executed before the instance is created. If the callback yields,
        the part after the yield will be executed after the creation.

        The callback should take one attribute that is the model instance.

        The callback can be freely named.

        >>> class Cowboy(sheraf.Model):
        ...     table = "cowboy_cb_creation"
        ...     name = sheraf.StringAttribute()
        ...
        >>>
        ... @Cowboy.on_creation
        ... def welcome_new_cowboys(cowboy):
        ...     yield
        ...     print(f"Welcome {cowboy.name}!")
        ...
        >>> with sheraf.connection():
        ...     george = Cowboy.create(name="George Abitbol")
        Welcome George Abitbol!
        """

        def wrapper(func):
            cls.cb_creation.append(func)
            return func

        return wrapper if not args else wrapper(args[0])

    @classmethod
    def on_deletion(cls, *args, **kwargs):
        """
        Decorator for callbacks to call on an instance deletion. The callback
        will be executed before the instance is deleted. If the callback yields,
        the part after the yield will be executed after the creation.

        The callback should take one attribute that is the model instance.

        The callback can be freely named.

        >>> class Cowboy(sheraf.Model):
        ...     table = "cowboy_cb_creation"
        ...     name = sheraf.StringAttribute()
        ...
        >>>
        ... @Cowboy.on_deletion
        ... def goodby_old_cowboys(cowboy):
        ...     print(f"So long {cowboy.name}!")
        ...
        >>> with sheraf.connection():
        ...     george = Cowboy.create(name="George Abitbol")
        ...     george.delete()
        So long George Abitbol!
        """

        def wrapper(func):
            cls.cb_deletion.append(func)
            return func

        return wrapper if not args else wrapper(args[0])

    @classmethod
    def _decorate(cls, mapping):
        instance = cls()
        instance.mapping = mapping
        return instance

    @classmethod
    def attribute_id(cls, name, attribute):
        raise NotImplementedError

    def __setattr__(self, name, value):
        if name not in self.attributes:
            super().__setattr__(name, value)
            return

        value = self.attributes[name].write(self, value)
        if self.attributes[name].write_memoization:
            super().__setattr__(name, value)

    def __delattr__(self, name):
        if name in self.attributes:
            self.attributes[name].delete(self)
            try:
                super().__delattr__(name)
            except AttributeError:
                return
        else:
            super().__delattr__(name)

    def __getattribute__(self, name):
        # TODO: Find a way to check that self.name exists or not without
        # using try/except AttributeError.
        # Because if a @property of this class, called `name` also raises
        # an AttributeError, actually we cannot say from where the AttributeError
        # was emitted.

        try:
            attribute = super().__getattribute__(name)
            if not isinstance(attribute, sheraf.attributes.Attribute):
                return attribute

        except AttributeError as exc:
            if f"object has no attribute '{name}'" not in str(exc):
                raise

        try:
            attribute = self.attributes[name]
        except KeyError:
            raise AttributeError(name)

        value = attribute.read(self)

        if self.attributes[name].read_memoization:
            super().__setattr__(name, value)

        return value

    def copy(self, **kwargs):
        r"""
        :param \*\*kwargs: Keywords arguments will be passed to
                           :func:`~sheraf.models.BaseModel.create` and thus
                           wont be copied.

        :return: a copy of this instance.
        """
        copy = self.__class__.create(**kwargs)
        for name, attr in self.attributes.items():
            if attr.is_created(self) and name not in kwargs:
                setattr(copy, name, getattr(self, name))
        return copy

    def keys(self):
        """
        :return: The instance attribute names.
        """
        return self.attributes.keys()

    def items(self):
        return (
            (key, self.attributes[key].read(self)) for key in self.attributes.keys()
        )

    def update(self, **kwargs):
        """Takes an arbitrary number of keywords arguments, and updates the
        instance attributes matching the arguments.

        This functions recursively calls :func:`sheraf.attributes.Attribute.edit` with `addition` and `edition` to `True`.

        >>> class Horse(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     horse = sheraf.InlineModelAttribute(Horse)
        ...
        >>> with sheraf.connection(commit=True):
        ...     george = Cowboy.create(name="George", horse={"name": "Centaurus"})
        ...     george.update(name="*incognito*", horse={"name": "Jolly Jumper"})
        ...     george.name
        '*incognito*'
        >>> with sheraf.connection():
        ...     george.horse.name
        'Jolly Jumper'

        Note that sub-instances are also edited.
        """
        self.edit(
            kwargs, addition=True, edition=True, deletion=False, replacement=False
        )

    def assign(self, **kwargs):
        """Takes an arbitrary number of keywords arguments, and updates the
        instance attributes matching the arguments.

        This functions recursively calls :func:`sheraf.attributes.Attribute.edit` with `addition`, `edition` and `deletion` to `True`.

        >>> class Arm(sheraf.InlineModel):
        ...     name = sheraf.SimpleAttribute()
        ...
        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     arms = sheraf.SmallListAttribute(sheraf.InlineModelAttribute(Arm))
        ...
        >>> with sheraf.connection(commit=True):
        ...     george = Cowboy.create(name="Three arms cowboy", arms=[
        ...         {"name": "Arm 1"}, {"name": "Arm 2"}, {"name": "Arm 3"},
        ...     ])
        ...     len(george.arms)
        3
        >>> with sheraf.connection(commit=True):
        ...     len(george.assign(name="George Abitbol", arms=[
        ...         {"name": "Superarm 1"}, {"name": "Superarm 2"},
        ...     ]).arms)
        2
        >>> with sheraf.connection():
        ...     george.arms[0].name
        'Superarm 1'

        George passed from 3 arms to only 2 because *assign* does remove sub instances. If we had called :func:`~sheraf.models.base.BaseModel.update` instead,
        George would have his two first arms be renamed *superarms* but, the third one would not have been removed.
        """
        return self.edit(
            kwargs, addition=True, edition=True, deletion=True, replacement=False
        )

    def edit(
        self,
        value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
        strict=False,
    ):
        """Take a dictionary and a set of options, and try to applies the
        dictionary values to the instance structure.

        :param value: The dictionary containing the values. The dictionary elements that
                      do not match the instance attributes will be ignored.
        :param addition: If *True*, elements present in *value* and absent from the instance attributes will be added.
        :param edition: If *True*, elements present in both *value* and the instance will be updated.
        :param deletion: If *True*, elements present in the instance and absent from *value* will be deleted.
        :param replacement: Like *edition*, but create a new element instead of updating one.
        :param strict: If strict is *True*, every keys in value must be sheraf attributes of the current model. Default is *False*.
        """
        for attr, new_value in value.items():
            try:
                old_value = self.attributes[attr].read(self)
                updated = self.attributes[attr].update(
                    old_value, new_value, addition, edition, deletion, replacement
                )
                self.__setattr__(attr, updated)
            except KeyError:
                if strict is True:
                    raise TypeError(
                        "TypeError: edit() got an unexpected keyword argument '{}'".format(
                            attr
                        )
                    )
        return self

    def save(self):
        for attr in self.attributes.values():
            attr.save(self)
        return self

    def reset(self, attribute):
        """
        Resets an attribute to its default value.
        """
        self.__setattr__(attribute, self.attributes[attribute].create(self))

    def __ne__(self, other):
        return not self == other

    def __eq__(self, other):
        return (
            hasattr(self, "mapping")
            and hasattr(other, "mapping")
            and self.mapping == other.mapping
        )

    def __getitem__(self, key):
        return self.attributes[key].read(self)

    def __setitem__(self, key, value):
        value = self.attributes[key].write(self, value)
        super().__setattr__(key, value)

    def __contains__(self, key):
        return key in self.attributes

    def __repr__(self):
        return f"<{self.__class__.__name__}>"



import random
import sys
import uuid

from sheraf.attributes.simples import IntegerAttribute
from sheraf.attributes.simples import StringUUIDAttribute

from .attributes import DatedNamedAttributesModel
from .attributes import IntAttributesModel
from .attributes import NamedAttributesModel
from .indexation import IndexedModel
from .indexation import SimpleIndexedModel


class UUIDIndexedModel:
    """Model using uuid4 as ids. Ids are stored as strings.

    >>> class MyUUIDModel(sheraf.IntIndexedModel):
    ...     table = "my_uuid_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    "e4bb714e-b5a8-40d6-bb69-ab3b932fbfe0"
    """

    id = StringUUIDAttribute(default=lambda: str(uuid.uuid4())).index(primary=True)


class IntIndexedModel:
    """Model using integers as ids.

    By default ids are 64bits integers.

    >>> class MyIntModel(sheraf.IntIndexedModel):
    ...     table = "my_int_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    383428472384721983
    """

    MAX_INT = sys.maxsize

    id = IntegerAttribute(default=lambda m: random.randint(0, m.MAX_INT)).index(
        primary=True
    )


class IntOrderedIndexedModel:
    """Model using increasing integers as ids.


    >>> class MyIntModel(sheraf.IntOrderedIndexedModel):
    ...     table = "my_int_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    ...     MyIntModel.create().id
    0
    1
    """

    id = IntegerAttribute(default=lambda m: m.count()).index(primary=True)


class IntIndexedNamedAttributesModel(
    NamedAttributesModel, IntIndexedModel, IndexedModel
):
    """The ids of this model are integers, and attributes are named."""


class IntOrderedNamedAttributesModel(
    NamedAttributesModel, IntOrderedIndexedModel, IndexedModel
):
    """The ids are 64bits integers, distributed ascendently starting at 0."""


class UUIDIndexedNamedAttributesModel(
    NamedAttributesModel, UUIDIndexedModel, IndexedModel
):
    """The ids of this model are UUID4, and attributes are named."""


class UUIDIndexedDatedNamedAttributesModel(
    DatedNamedAttributesModel, UUIDIndexedModel, IndexedModel
):
    """The ids of this model are UUID4, the attributes are named, and any
    modification on the model will update its modification datetime."""


class IntIndexedIntAttributesModel(IntAttributesModel, IntIndexedModel, IndexedModel):
    """The ids of this models are integers, and the ids of its attributes are
    also integers."""


class AttributeModel(NamedAttributesModel, SimpleIndexedModel):
    """
    This model is expected to be used with :class:`~sheraf.attributes.models.IndexedModelAttribute`.
    Its usage is mainly the same as any :class:`~sheraf.models.indexation.BaseIndexedModel`.
    """


Model = UUIDIndexedDatedNamedAttributesModel



from rich.console import Console
from rich.markdown import Markdown
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import TimeRemainingColumn
from rich.table import Table
from rich.text import Text
from sheraf.health.utils import discover_models

from .checks import check_attributes_index
from .checks import check_conflict_resolution
from .checks import check_model_index


# Set the list of check functions to be run
INSTANCE_CHECK_FUNCS = {}
ATTRIBUTE_CHECK_FUNCS = {
    "index": check_attributes_index,
}
MODEL_CHECK_FUNCS = {"index": check_model_index}
OTHER_CHECK_FUNCS = {"check_conflict_resolution": check_conflict_resolution}


def check_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None,
    console=None,
):
    """
    Takes some modules in parameters.

    :param model_checks: If None, check also model consistency (see constant model_check_funcs)
    :param instance_checks: If None, check all instance consistency rules. Else, give the list of wanted
        checking rules (see constant instance_check_funcs)
    :param attribute_checks: If None, check all attribute consistency rules. Else, give the list of wanted
        checking rules (see constant instance_check_funcs)
        The function will discover models in the modules, analyze every
        model instance, and return a health report in JSON. Depending on the
        result, you may need to apply some fixes.
    """
    models = discover_models(*args)
    health_report = {}

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    for check_key in other_checks:
        health_report[check_key] = OTHER_CHECK_FUNCS[check_key]()

    with Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        TextColumn("{task.completed}", justify="right"),
        TextColumn("/"),
        TextColumn("{task.total}"),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeRemainingColumn(),
        console=console,
        transient=True,
    ) as progress:
        tasks = {}
        for model_path, model in models:
            tasks[model] = progress.add_task(model.__name__)

        for model_path, model in models:
            progress.update(tasks[model], total=model.count())

        for model_path, model in models:
            progress.start_task(tasks[model])

            for check_key in instance_checks:
                instance_check_func = INSTANCE_CHECK_FUNCS[check_key]
                health_report.setdefault(instance_check_func.__name__, {}).setdefault(
                    model_path, {"ok": 0, "ko": 0}
                )

            for check_key in model_checks:
                model_check_func = MODEL_CHECK_FUNCS[check_key]
                health_report.setdefault(model_check_func.__name__, {})[
                    model_path
                ] = model_check_func(model)

            # Iterate on instances
            for m in model.all():
                progress.update(tasks[model], advance=1)
                for check_key in instance_checks:
                    check_func = INSTANCE_CHECK_FUNCS[check_key]
                    model_instance_result = health_report[check_func.__name__][
                        model_path
                    ]
                    if check_func(m):
                        model_instance_result["ok"] += 1
                    else:
                        model_instance_result["ko"] += 1

                for check_key in attribute_checks:
                    check_func = ATTRIBUTE_CHECK_FUNCS[check_key]
                    health_report.setdefault(check_func.__name__, {})
                    for attribute_name, bool_value in check_func(m).items():
                        attribute_result = (
                            health_report[check_func.__name__]
                            .setdefault(model_path, {})
                            .setdefault(attribute_name, {"ok": 0, "ko": 0})
                        )
                        if bool_value:
                            attribute_result["ok"] += 1
                        else:
                            attribute_result["ko"] += 1
    return health_report


def _print_check_other_health_result(console, check_reason, health_table, help):
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=check_reason,
        expand=True,
        caption=help,
    )
    table.add_column("Check")
    table.add_column("State")

    table.add_row(check_reason, "OK" if health_table[check_reason] else "KO")
    console.print(table)


def _print_check_model_health_result(console, check_reason, health_table, help):
    table_key = "check_model_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")

    for model_path, attributes in health_table.get(table_key, {}).items():
        table.add_row(
            model_path,
            str(sum(values["ko"] for values in attributes.values())),
            str(sum(values["ok"] for values in attributes.values())),
        )

        for attribute_name, values in attributes.items():
            table.add_row(f"  - {attribute_name}", str(values["ko"]), str(values["ok"]))

    if health_table.get(table_key):
        console.print(table)
    else:
        console.print("  No model to visit.")


def _print_check_instance_health_result(console, check_reason, health_table, help):
    """
    :param check_reason: one among model_checks keys
    :param health_table: result of a check function
    """
    table_key = "check_instance_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")
    for model_path, values in health_table.get(table_key, {}).items():
        table.add_row(model_path, str(values["ko"]), str(values["ok"]))
    console.print(table)


def _print_check_attribute_health_result(console, check_reason, health_table, help):
    table_key = "check_attributes_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")

    for model_path, attributes in health_table.get(table_key, {}).items():
        table.add_row(
            model_path,
            str(sum(values["ko"] for values in attributes.values())),
            str(sum(values["ko"] for values in attributes.values())),
        )

        for attribute_name, values in attributes.items():
            table.add_row(f"  - {attribute_name}", str(values["ko"]), str(values["ok"]))

    if health_table.get(table_key):
        console.print(table)
    else:
        console.print("  No model to visit.")


def print_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None,
):
    console = Console(width=100)
    """Takes some modules in parameters (e.g. "american._class.cowboy_module").

    The function will discover models in the modules, analyze every model instance, and return
    a health report in a human readable format. Depending on the result, you may need to apply some fixes.

    This function does not edit any data and is safe to be executed in a production shell.
    """
    console.print(
        Text(
            "====== _ ==================  __ ==\n"
            "=     | |                   / _| =\n"
            "=  ___| |__   ___ _ __ __ _| |_  =\n"
            "= / __| '_ \\ / _ \\ '__/ _` |  _| =\n"
            "= \\__ \\ | | |  __/ | | (_| | |   =\n"
            "= |___/_| |_|\\___|_|  \\__,_|_|   =\n"
            "==================================",
            justify="center",
        )
    )

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    health = check_health(
        *args,
        model_checks=model_checks,
        instance_checks=instance_checks,
        attribute_checks=attribute_checks,
        other_checks=other_checks,
        console=console,
    )

    for other_check_type in other_checks:
        console.print()
        _print_check_other_health_result(
            console,
            other_check_type,
            health,
            Markdown(OTHER_CHECK_FUNCS[other_check_type].__doc__),
        )
        console.print()

    for model_check_type in model_checks:
        console.print()
        _print_check_model_health_result(
            console,
            model_check_type,
            health,
            Markdown(MODEL_CHECK_FUNCS[model_check_type].__doc__),
        )
        console.print()

    for instance_check_type in instance_checks:
        console.print()
        _print_check_instance_health_result(
            console,
            instance_check_type,
            health,
            Markdown(INSTANCE_CHECK_FUNCS[instance_check_type].__doc__),
        )
        console.print()

    for attribute_check_type in attribute_checks:
        console.print()
        _print_check_attribute_health_result(
            console,
            attribute_check_type,
            health,
            Markdown(ATTRIBUTE_CHECK_FUNCS[attribute_check_type].__doc__),
        )
        console.print()
